using System;

namespace NdisMonitor
{
	//
	// ## ATTRIBUTES ##
	//

	[ AttributeUsage( AttributeTargets.Class, AllowMultiple = false, Inherited = false ) ]
	public class PartInfoAttribute : System.Attribute
	{
		// construction.
		public PartInfoAttribute( string name )
		{
			this._name = name;
		}

		// data.
		protected string		_name = "";

		// public properties.
		public string Name
		{
			get
			{
				return _name;
			}
		}
	}

	public delegate string PrintFnType ();
	public delegate int FieldSizeFnType ();

	[ AttributeUsage( AttributeTargets.Field, AllowMultiple = false, Inherited = false ) ]
	public class FieldInfoAttribute : System.Attribute
	{
		// construction.
		public FieldInfoAttribute( string name, int offset )
		{
			this._name = name;
			this._offset = offset;
		}

		// data.
		protected string		_name = "";
		protected int			_offset = 0;
		protected int			_size = -1; // -1 means autodetect through reflection.
		protected string		_printFn = null;
		protected string		_fieldSizeFn = null;

		// public properties.
		public string Name
		{
			get
			{
				return _name;
			}
		}
		public int Offset
		{
			get
			{
				return _offset;
			}
		}
		public int Size
		{
			get
			{
				return _size;
			}
			set
			{
				_size = value;
			}
		}
		public string PrintFn
		{
			get
			{
				return _printFn;
			}
			set
			{
				_printFn = value;
			}
		}
		public string FieldSizeFn
		{
			get
			{
				return _fieldSizeFn;
			}
			set
			{
				_fieldSizeFn = value;
			}
		}
	}

	//
	// ## BASE CLASSES ##
	//

	public abstract class Packet_Part
	{
		// construction.

		protected Packet_Part( int defLength )
		{
			_length = defLength;
		}

		// data.

		public int			_length = 0;
	}

	public class RawPacket
	{
		//
		// # construction. #
		//

		protected RawPacket( byte[] data )
		{
			_data = data;
		}

		//
		// # data. #
		//

		public byte[]		_data;

		//
		// # methods. #
		//

		public static RawPacket GET( byte[] data )
		{
			EthernetPacket		ep = EthernetPacket.GET( data );
			if ( ep == null )
				return new RawPacket( data );
			else
				return ep;
		}

		//
		// # overrideables #
		//

		public virtual Packet_Part[] GetParts ()
		{
			return null;
		}

		//
		// # helpers. #
		//

		public static UInt16 ReadUInt16( byte[] byteArray, int pos )
		{
			return (UInt16) ( ( byteArray[ pos ] << 8 ) | byteArray[ pos + 1 ] );
		}

		public static UInt32 ReadUInt32( byte[] byteArray, int pos )
		{
			return (UInt32) (
				( byteArray[ pos + 0 ] << 24 ) |
				( byteArray[ pos + 1 ] << 16 ) |
				( byteArray[ pos + 2 ] << 8 ) |
				( byteArray[ pos + 3 ] << 0 )
				);
		}

		public static int GetTotalLength( Packet_Part[] parts, Packet_Part basePart, bool raiseExIfZero )
		{
			if ( parts == null )
			{
				if ( raiseExIfZero )
					throw new Exception ();
				return 0;
			}

			int			retval = 0;

			foreach( Packet_Part p in parts )
			{
				if ( p == basePart )
					break;
				retval += p._length;
			}

			if ( retval == 0 && raiseExIfZero )
				throw new Exception ();
			return retval;
		}
	}

	//
	// ## ETHERNET PACKET ##
	//

	public class EthernetPacket : RawPacket
	{
		// construction.

		protected EthernetPacket( byte[] data ) : base( data )
		{
			_macHeader = Ethernet_Base.GET( data );
			_netHeader = Network_Protocol.GET( this, _macHeader );
			_tranHeader = Transport_Protocol.GET( this, new Packet_Part[ 2 ] { _macHeader, _netHeader } );
		}

		// data.

		public Ethernet_Base			_macHeader = null;
		public Network_Protocol			_netHeader = null;
		public Transport_Protocol		_tranHeader = null;
		public Packet_Part[]			_packetData = null;

		// methods.

		public static new EthernetPacket GET( byte[] data )
		{
			EthernetPacket			ep = new EthernetPacket( data );
			if ( ep._macHeader == null || ep._netHeader == null || ep._tranHeader == null )
				return null;
			else
				return ep;
		}

		public override Packet_Part[] GetParts ()
		{
			int				size = 3 + ( _packetData == null ? 0 : _packetData.Length );

			Packet_Part[]	retval = new Packet_Part[ size ];
			retval[ 0 ] = _macHeader;
			retval[ 1 ] = _netHeader;
			retval[ 2 ] = _tranHeader;
			for( int i=0; i<size-3; i ++ )
				retval[ 3 + i ] = _packetData[ i ];

			return retval;
		}
	}

	//
	// ## FRAME TYPES ##
	//

	[ PartInfo( "Ethernet base header" ) ]
	public abstract class Ethernet_Base : Packet_Part
	{
		// construction.

		protected Ethernet_Base( byte[] data ) : base( 12 )
		{
			if ( data == null || data.Length < _length )
				throw new Exception();

			for ( int i=0; i<6; i ++ )
			{
				_macDst[ i ] = data[ i ];
				_macSrc[ i ] = data[ i + 6 ];
			}
		}

		// data.

		[ FieldInfo( "MAC Dest Address", 0, PrintFn = "MacDstToString" ) ]
		public byte[]			_macDst = new byte[ 6 ];
		[ FieldInfo( "MAC Src Address", 6, PrintFn = "MacSrcToString" ) ]
		public byte[]			_macSrc = new byte[ 6 ];

		// methods.

		public static Ethernet_Base GET( byte[] data )
		{
			if ( data == null || data.Length < 14 )
				return null;

			UInt16			_ctrl = RawPacket.ReadUInt16( data, 12 );

			try
			{
				if ( _ctrl <= 1500 ) // 802 format.
				{
					if ( data.Length >= 22 &&
						data[ 14 ] == 0xAA &&
						data[ 15 ] == 0xAA &&
						data[ 16 ] == 0x03 ) // SNAP format.
					{
						return new Ethernet_SNAP( data );
					}
					else // 802 format.
					{
						return new Ethernet_802_2( data );
					}
				}
				else // dix format.
				{
					return new Ethernet_II( data );
				}
			}
			catch
			{
				return null;
			}
		}

		protected string MacAddressToString( byte[] addr )
		{
			if ( addr == null || addr.Length != 6 )
				throw new Exception ();

			string		retval = "";
			for( int i=0; i<6; i ++ )
			{
				string			s = addr[ i ].ToString( "X" );
				if ( s.Length == 1 )
					s = "0" + s;
				retval += s;
			}

			return retval;
		}

		public string MacDstToString ()
		{
			return MacAddressToString( _macDst );
		}

		public string MacSrcToString ()
		{
			return MacAddressToString( _macSrc );
		}
	}

	[ PartInfo( "Ethernet II header" ) ]
	public class Ethernet_II : Ethernet_Base
	{
		// construction.

		internal Ethernet_II( byte[] data ) : base( data )
		{
			_length += 2;

			if ( data == null || data.Length < _length )
				throw new Exception();

			_eType = RawPacket.ReadUInt16( data, 12 );
		}

		// data.

		[ FieldInfo( "Ethertype", 12 ) ]
		public UInt16			_eType = 0;
	}

	[ PartInfo( "Ethernet 802.3 header" ) ]
	public class Ethernet_802_3 : Ethernet_Base
	{
		// construction.

		internal Ethernet_802_3( byte[] data ) : base( data )
		{
			_length += 2;

			if ( data == null || data.Length < _length )
				throw new Exception();

			_totPacketLen = RawPacket.ReadUInt16( data, 12 );
		}

		// data.

		[ FieldInfo( "Total Packet Len", 12 ) ]
		public UInt16			_totPacketLen = 0;
	}

	[ PartInfo( "Ethernet 802.2 header" ) ]
	public class Ethernet_802_2 : Ethernet_802_3
	{
		// construction.

		internal Ethernet_802_2( byte[] data ) : base( data )
		{
			_length += 3;

			if ( data == null || data.Length < _length )
				throw new Exception();

			_dSap = data[ 14 ];
			_sSap = data[ 15 ];
			_control = data[ 16 ];
		}

		// data.

		[ FieldInfo( "DSAP", 14 ) ]
		public Byte			_dSap = 0;
		[ FieldInfo( "SSAP", 15 ) ]
		public Byte			_sSap = 0;
		[ FieldInfo( "Control", 16 ) ]
		public Byte			_control = 0;
	}

	[ PartInfo( "Ethernet SNAP header" ) ]
	public class Ethernet_SNAP : Ethernet_802_2
	{
		// construction.

		internal Ethernet_SNAP( byte[] data ) : base( data )
		{
			_length += 5;

			if ( data == null || data.Length < _length )
				throw new Exception();

			if ( _dSap != 0xAA ||
				_sSap != 0xAA ||
				_control != 0x03 )
				throw new Exception();

			for( int i=0; i<3; i ++ )
				_snapSubHdrStart[ i ] = data[ i + 17 ];
			_ethernetIIType = RawPacket.ReadUInt16( data, 20 );
		}

		// data.

		[ FieldInfo( "Initial Bytes", 17 ) ]
		public byte[]			_snapSubHdrStart = new byte[ 3 ];
		[ FieldInfo( "Ethertype", 20 ) ]
		public UInt16			_ethernetIIType = 0;
	}

	//
	// # NETWORK PROTOCOLS #
	//

	public abstract class Network_Protocol : Packet_Part
	{
		// construction.

		protected Network_Protocol () : base( 0 )
		{ }

		// methods.

		public static Network_Protocol GET( RawPacket rawPacket, Ethernet_Base macHeader )
		{
			if ( rawPacket == null || macHeader == null || rawPacket._data == null )
				return null;

			UInt16			eType = 0;

			if ( macHeader is Ethernet_II )
				eType = ( (Ethernet_II) macHeader )._eType;
			else if ( macHeader is Ethernet_SNAP )
				eType = ( (Ethernet_SNAP) macHeader )._ethernetIIType;
			else
				return new NetProt_Undefined ();

			switch( eType )
			{
				case 0x0800: // IP, Internet Protocol.
				{
					NetProt_IP		ip = NetProt_IP.GET( rawPacket, macHeader );
					if ( ip == null )
						return new NetProt_Undefined ();
					else
						return ip;
				}

				default: // ???.
					return new NetProt_Undefined ();
			}
		}
	}

	public class NetProt_Undefined : Network_Protocol
	{
		// ### NOTHING HERE ###
	}

	public enum EIpVersion
	{
		IP = 4,		// IP, Internet Protocol.
		ST = 5,		// ST, ST Datagram Mode.
		SIP = 6,	// SIP, Simple Internet Protocol.
		TPIX = 7,	// TP/IX, The Next Internet.
		PIP = 8,	// PIP, The P Internet Protocol.
		TUBA = 9	// TUBA.
	}

	public enum EIpTosPrecedence
	{
		Routine = 0,
		Priority = 1,
		Immediate = 2,
		Flash = 3,
		Flash_override = 4,
		CRITIC_ECP = 5,
		Internetwork_control = 6,
		Network_control = 7
	}

	public enum EIpProtocol
	{
		HOPOPT = 0, // IPv6 Hop-by-Hop Option
		ICMP = 1, // Internet Control Message Protocol
		IGAP = 2, // IGMP for user Authentication Protocol
		GGP = 3, // Gateway to Gateway Protocol
		IP_in_IP_encap = 4, // IP in IP encapsulation
		IST = 5, // Internet Stream Protocol
		TCP = 6, // Transmission Control Protocol
		UCL = 7, // ???
		EGP = 8, // Exterior Gateway Protocol
		IGRP = 9, // Interior Gateway Routing Protocol
		BBN_RCC_Monit = 10, // BBN RCC Monitoring
		NVP = 11, // Network Voice Protocol
		PUP = 12, // ???
		ARGUS = 13, // ???
		EMCON = 14, // Emission Control Protocol
		XNET = 15, // Cross Net Debugger
		Chaos = 16, // ???
		UDP = 17, // User Datagram Protocol
		TMux = 18, // Transport Multiplexing Protocol
		DCN_Mea_Subs = 19, // DCN Measurement Subsystems
		HMP = 20, // Host Monitoring Protocol
		PRM = 21, // Packet Radio Measurement
		Xerox_NS_IDP = 22, // XEROX NS IDP
		Trunk_1 = 23, // ???
		Trunk_2 = 24, // ???
		Leaf_1 = 25, // ???
		Leaf_2 = 26, // ???
		RDP = 27, // Reliable Data Protocol
		IRTP = 28, // Internet Reliable Transaction Protocol
		ISO_TPC4 = 29, // ISO Transport Protocol Class 4
		NETBLT = 30, // Network Block Transfer
		MFE = 31, // MFE Network Services Protocol
		MERIT = 32, // MERIT Internodal Protocol
		SEP = 33, // Sequential Exchange Protocol
		TPCP = 34, // Third Party Connect Protocol
		IDPR = 35, // Inter-Domain Policy Routing Protocol
		XTP = 36, // Xpress Transfer Protocol
		DDP = 37, // Datagram Delivery Protocol
		CMTP = 38, // Control Message Transport Protocol
		TPPP = 39, // TP++ Transport Protocol
		IL = 40, // IL Transport Protocol
		IPv6_over_IPv4 = 41, // IPv6 over IPv4
		SDRP = 42, // Source Demand Routing Protocol
		IPv6RH = 43, // IPv6 Routing header
		IPv6FH = 44, // IPv6 Fragment header
		IDRP = 45, // Inter-Domain Routing Protocol
		RSVP = 46, // Reservation Protocol
		GRE = 47, // General Routing Encapsulation
		MHRP = 48, // Mobile Host Routing Protocol
		BNA = 49, // ???
		ESP = 50, // Encapsulating Security Payload
		AH = 51, // Authentication Header
		INLSTUBA = 52, // Integrated Net Layer Security TUBA
		IPEnc = 53, // IP with Encryption
		NARP = 54, // NBMA Address Resolution Protocol
		MEP = 55, // Minimal Encapsulation Protocol
		TLSP = 56, // Transport Layer Security Protocol using Kryptonet key management
		SKIP = 57, // ???
		ICMPv6 = 58, // Internet Control Message Protocol for IPv
		IPv6_No_Next_Head = 59, // IPv6 No Next Header
		IPv6DestOpt = 60, // Destination Options for IPv6
		AnyHost = 61, // Any host internal protocol
		CFTP = 62, // ???
		AnyLocal = 63, // Any local network
		EXPAK = 64, // SATNET and Backroom EXPAK
		Kryptolan = 65, // ???
		MIT_Remote_VDisk = 66, // MIT Remote Virtual Disk Protocol
		Internet_Pluribus = 67, // Internet Pluribus Packet Core
		AnyDistrFS = 68, // Any distributed file system
		SATNET_Monitor = 69, // SATNET Monitoring
		VISA = 70, // VISA Protocol
		Internet_Packet = 71, // Internet Packet Core Utility
		Net_Exe = 72, // Computer Protocol Network Executive
		Heart_Beat = 73, // Computer Protocol Heart Beat
		Wang_Span = 74, // Wang Span Network
		Packet_Video = 75, // Packet Video Protocol
		Backr_SATNET_Monitor = 76, // Backroom SATNET Monitoring
		SUN_ND_temp = 77, // SUN ND PROTOCOL-Temporary
		WIDEBAND_Monitor = 78, // WIDEBAND Monitoring
		WIDEBAND_EXPAK = 79, // WIDEBAND EXPAK
		ISO_IP = 80, // ISO-IP
		VMTP = 81, // Versatile Message Transaction Protocol
		SECURE_VMT = 82, // ???
		VINES = 83, // ???
		TTP = 84, // ???
		NSFNET_IGP = 85, // ???
		DGP = 86, // Dissimilar Gateway Protocol
		TCF = 87, // ???
		EIGRP = 88, // ???
		OSPF = 89, // Open Shortest Path First Routing Protocol
		SRPCP = 90, // Sprite RPC Protocol
		LARP = 91, // Locus Address Resolution Protocol
		MTP = 92, // Multicast Transport Protocol
		AX25 = 93, // AX.25
		IP_in_IP = 94, // IP-within-IP Encapsulation Protocol
		MICP = 95, // Mobile Internetworking Control Protocol
		Semaphore_Com = 96, // Semaphore Communications Sec. Pro
		EtherIP = 97, // ???
		Encap_Head = 98, // Encapsulation Header
		AnyPrivateEnc = 99, // Any private encryption scheme
		GMTP = 100, // ???
		IFMP = 101, // Ipsilon Flow Management Protocol
		PNNI_over_IP = 102, // ???
		PIM = 103, // Protocol Independent Multicast
		ARIS = 104, // ???
		SCPS = 105, // ???
		QNX = 106, // ???
		AN = 107, // Active Networks
		IPPCP = 108, // IP Payload Compression Protocol
		SNP = 109, // Sitara Networks Protocol
		Compaq_Peer = 110, // Compaq Peer Protocol
		IPX_in_IP = 111, // IPX in IP
		VRRP = 112, // Virtual Router Redundancy Protocol
		PGM = 113, // Pragmatic General Multicast
		AnyZeroHop = 114, // Any 0-hop protocol
		L2TP = 115, // Level 2 Tunneling Protocol
		DDX = 116, // D-II Data Exchange
		IATP = 117, // Interactive Agent Transfer Protocol
		ST = 118, // Schedule Transfer
		SRP = 119, // SpectraLink Radio Protocol
		UTI = 120, // ???
		SMP = 121, // Simple Message Protocol
		SM = 122, // ???
		PTP = 123, // Performance Transparency Protocol
		ISIS_over_IPv4 = 124, // ISIS over IPv4
		FIRE = 125, // ???
		CRTP = 126, // Combat Radio Transport Protocol
		CRUDP = 127, // Combat Radio User Datagram
		SSCOPMCE = 128, // ???
		IPLT = 129, // ???
		SPS = 130, // Secure Packet Shield
		PIPE = 131, // Private IP Encapsulation within IP
		SCTP = 132, // Stream Control Transmission Protocol
		Fibre = 133, // Fibre Channel
		RSVP_E2E_IGN = 134, // RSVP-E2E-IGNORE
		Mobility_Head = 135, // Mobility Header
		UDP_Lite = 136, // Lightweight User Datagram Protocol
		MPLS_in_IP = 137 // MPLS-in-IP
	}

	[ PartInfo( "IP Network Protocol header" ) ]
	public class NetProt_IP : Network_Protocol
	{
		// construction.

		protected NetProt_IP( RawPacket rawPacket, Ethernet_Base macHeader ) : base ()
		{
			// fill the fields.

			int			pos = macHeader._length;

			byte		verIhl = rawPacket._data[ pos ++ ];
			_version = (EIpVersion) ( verIhl >> 4 );
			_ihl = (byte) ( verIhl & 0x0F );

			byte		tos = rawPacket._data[ pos ++ ];
			_tosPrec = (EIpTosPrecedence) ( tos >> 5 );
			_tosMinimizeDelay = ( tos & 0x10 ) != 0 ? true : false;
			_tosMaximizeThroughput = ( tos & 0x08 ) != 0 ? true : false;
			_tosMaximizeReliability = ( tos & 0x04 ) != 0 ? true : false;
			_tosMinimizeMonetaryCost = ( tos & 0x02 ) != 0 ? true : false;

			_totalLength = (ushort) RawPacket.ReadUInt16( rawPacket._data, pos );
			pos += 2;

			_identification = (ushort) RawPacket.ReadUInt16( rawPacket._data, pos );
			pos += 2;

			UInt16		flagsFragOff = RawPacket.ReadUInt16( rawPacket._data, pos );
			pos += 2;
			_flagsDontFragment = ( flagsFragOff & 0x4000 ) != 0 ? true : false;
			_flagsMoreFragments = ( flagsFragOff & 0x2000 ) != 0 ? true : false;
			_fragmentOffset = (ushort) ( flagsFragOff & 0x1FFF );

			_ttl = rawPacket._data[ pos ++ ];

			_protocol = (EIpProtocol) rawPacket._data[ pos ++ ];

			_hdrChecksum = (ushort) RawPacket.ReadUInt16( rawPacket._data, pos );
			pos += 2;

			_srcIP = (uint) RawPacket.ReadUInt32( rawPacket._data, pos );
			pos += 4;

			_dstIP = (uint) RawPacket.ReadUInt32( rawPacket._data, pos );
			pos += 4;

			// set length.

			_length = _ihl * 4;
		}

		// data.

		[ FieldInfo( "IP Version", 0, Size = 1 ) ]
		public EIpVersion		_version;
		[ FieldInfo( "IHL", 0 ) ]
		public byte				_ihl;
		[ FieldInfo( "Tos Precedence", 1, Size = 1 ) ]
		public EIpTosPrecedence	_tosPrec;
		[ FieldInfo( "Tos Minimize Delay", 1 ) ]
		public bool				_tosMinimizeDelay;
		[ FieldInfo( "Tos Maximize Throughput", 1 ) ]
		public bool				_tosMaximizeThroughput;
		[ FieldInfo( "Tos Maximize Reliability", 1 ) ]
		public bool				_tosMaximizeReliability;
		[ FieldInfo( "Tos Minimize Monetary Cost", 1 ) ]
		public bool				_tosMinimizeMonetaryCost;
		[ FieldInfo( "Total Length", 2 ) ]
		public ushort			_totalLength;
		[ FieldInfo( "Identification", 4 ) ]
		public ushort			_identification;
		[ FieldInfo( "Flags Dont Fragment", 6, Size = 1 ) ]
		public bool				_flagsDontFragment;
		[ FieldInfo( "Flags More Fragments", 6, Size = 1 ) ]
		public bool				_flagsMoreFragments;
		[ FieldInfo( "Fragment Offset", 6 ) ]
		public ushort			_fragmentOffset;
		[ FieldInfo( "TTL", 8 ) ]
		public byte				_ttl;
		[ FieldInfo( "Protocol", 9, Size = 1 ) ]
		public EIpProtocol		_protocol;
		[ FieldInfo( "Hdr Checksum", 10 ) ]
		public ushort			_hdrChecksum;
		[ FieldInfo( "Source IP", 12, PrintFn = "SrcIPToString" ) ]
		public uint				_srcIP;
		[ FieldInfo( "Destination IP", 16, PrintFn = "DstIPToString" ) ]
		public uint				_dstIP;

		// methods.

		public static new NetProt_IP GET( RawPacket rawPacket, Ethernet_Base macHeader )
		{
			try
			{
				NetProt_IP			ip = new NetProt_IP( rawPacket, macHeader );
				return ip;
			}
			catch
			{
				return null;
			}
		}

		protected string UIntToIPAddressString( uint ui )
		{
			string			retval = "";
			int				shift = 24;

			for( int i=0; i<4; i ++ )
			{
				byte		b = (byte) ( ( ui >> shift ) & 0xFF );
				retval += b.ToString () + ( i!=3 ? "." : "" );
				shift -= 8;
			}

			return retval;
		}

		public string SrcIPToString()
		{
			return UIntToIPAddressString( _srcIP );
		}

		public string DstIPToString ()
		{
			return UIntToIPAddressString( _dstIP );
		}
	}

	//
	// # TRANSPORT PROTOCOLS #
	//

	public abstract class Transport_Protocol : Packet_Part
	{
		// construction.

		protected Transport_Protocol () : base( 0 )
		{ }

		// methods.

		public static Transport_Protocol GET( RawPacket rawPacket, Packet_Part[] precParts )
		{
			if ( rawPacket == null || precParts == null || rawPacket._data == null || precParts.Length == 0 )
				return null;

			Packet_Part		lastPart = precParts[ precParts.Length - 1 ];
			if ( lastPart == null )
				return null;

			EIpProtocol			protocol;

			if ( lastPart is NetProt_IP )
				protocol = ( (NetProt_IP) lastPart )._protocol;
			else
				return new TranProt_Undefined ();

			Transport_Protocol			tp = null;

			switch( protocol )
			{
				case EIpProtocol.ICMP:
					tp = TranProt_ICMP.GET( rawPacket, precParts );
					break;

				case EIpProtocol.TCP:
					tp = TranProt_TCP.GET( rawPacket, precParts );
					break;

				case EIpProtocol.UDP:
					tp = TranProt_UDP.GET( rawPacket, precParts );
					break;

				default: // ???.
					return new TranProt_Undefined ();
			}

			if ( tp == null )
				tp = new TranProt_Undefined ();

			return tp;
		}
	}

	public class TranProt_Undefined : Transport_Protocol
	{
		// ### NOTHING HERE ###
	}

	public enum EIcmpType
	{
		echo_reply = 0,
		destination_unreachable = 3,
		source_quench = 4,
		redirect = 5,
		alternate_host_address = 6,
		echo_request = 8,
		router_advertisement = 9,
		router_solicitation = 10,
		time_exceeded = 11,
		parameter_problem = 12,
		timestamp_request = 13,
		timestamp_reply = 14,
		information_request = 15,
		information_reply = 16,
		address_mask_request = 17,
		address_mask_reply = 18,
		traceroute = 30,
		conversion_error = 31,
		mobile_host_redirect = 32,
		ipv6_where_are_you = 33,
		ipv6_i_am_here = 34,
		mobile_registration_request = 35,
		mobile_registration_reply = 36,
		domain_name_request = 37,
		domain_name_reply = 38,
		skip_algorithm_discovery_protocol = 39,
		photuris_security_failures = 40
	}

	[ PartInfo( "ICMP Transport Protocol header" ) ]
	public class TranProt_ICMP : Transport_Protocol
	{
		// construction.

		protected TranProt_ICMP( RawPacket rawPacket, Packet_Part[] precParts ) : base()
		{
			// fill the fields.

			int			pos = RawPacket.GetTotalLength( precParts, null, true );

			_type = (EIcmpType) rawPacket._data[ pos ++ ];

			_code = rawPacket._data[ pos ++ ];

			_hdrChecksum = (ushort) RawPacket.ReadUInt16( rawPacket._data, pos );
			pos += 2;

			// set length.

			_length = 4;
		}

		// data.

		[ FieldInfo( "Type", 0, Size = 1 ) ]
		public EIcmpType		_type;
		[ FieldInfo( "Code", 1 ) ]
		public byte				_code;
		[ FieldInfo( "Hdr Checksum", 2 ) ]
		public ushort			_hdrChecksum;

		// methods.

		public static new TranProt_ICMP GET( RawPacket rawPacket, Packet_Part[] precParts )
		{
			try
			{
				TranProt_ICMP		tp = new TranProt_ICMP( rawPacket, precParts );
				return tp;
			}
			catch
			{
				return null;
			}
		}
	}

	[ PartInfo( "TCP Transport Protocol header" ) ]
	public class TranProt_TCP : Transport_Protocol
	{
		// construction.

		protected TranProt_TCP( RawPacket rawPacket, Packet_Part[] precParts ) : base()
		{
			// fill the fields.

			int			pos = RawPacket.GetTotalLength( precParts, null, true );

			_srcPort = (ushort) RawPacket.ReadUInt16( rawPacket._data, pos );
			pos += 2;

			_dstPort = (ushort) RawPacket.ReadUInt16( rawPacket._data, pos );
			pos += 2;

			_sequenceNum = (uint) RawPacket.ReadUInt32( rawPacket._data, pos );
			pos += 4;

			_ackNum = (uint) RawPacket.ReadUInt32( rawPacket._data, pos );
			pos += 4;

			ushort		dataOffEcnCB = (ushort) RawPacket.ReadUInt16( rawPacket._data, pos );
			pos += 2;
			_dataOff = (byte) ( dataOffEcnCB >> 12 );
			_reserved = (byte) ( ( dataOffEcnCB >> 9 ) & 0x07 );
			_ecnNS = ( ( dataOffEcnCB >> 8 ) & 1 ) == 0 ? false : true;
			_ecnCWR = ( ( dataOffEcnCB >> 7 ) & 1 ) == 0 ? false : true;
			_ecnECE = ( ( dataOffEcnCB >> 6 ) & 1 ) == 0 ? false : true;
			_controlURG = ( ( dataOffEcnCB >> 5 ) & 1 ) == 0 ? false : true;
			_controlACK = ( ( dataOffEcnCB >> 4 ) & 1 ) == 0 ? false : true;
			_controlPSH = ( ( dataOffEcnCB >> 3 ) & 1 ) == 0 ? false : true;
			_controlRST = ( ( dataOffEcnCB >> 2 ) & 1 ) == 0 ? false : true;
			_controlSYN = ( ( dataOffEcnCB >> 1 ) & 1 ) == 0 ? false : true;
			_controlFIN = ( ( dataOffEcnCB >> 0 ) & 1 ) == 0 ? false : true;

			_window = (ushort) RawPacket.ReadUInt16( rawPacket._data, pos );
			pos += 2;

			_checkSum = (ushort) RawPacket.ReadUInt16( rawPacket._data, pos );
			pos += 2;

			_urgentPtr = (ushort) RawPacket.ReadUInt16( rawPacket._data, pos );
			pos += 2;

			// set length.

			_length = _dataOff * 4;
		}

		// data.

		[ FieldInfo( "Source Port", 0 ) ]
		public ushort			_srcPort;
		[ FieldInfo( "Destination Port", 2 ) ]
		public ushort			_dstPort;
		[ FieldInfo( "Sequence Number", 4 ) ]
		public uint				_sequenceNum;
		[ FieldInfo( "ACK Number", 8 ) ]
		public uint				_ackNum;
		[ FieldInfo( "Data Offset", 12 ) ]
		public byte				_dataOff;
		[ FieldInfo( "Reserved", 12 ) ]
		public byte				_reserved;
		[ FieldInfo( "ECN NS", 12 ) ]
		public bool				_ecnNS;
		[ FieldInfo( "ECN CWR", 13 ) ]
		public bool				_ecnCWR;
		[ FieldInfo( "ECN ECE", 13 ) ]
		public bool				_ecnECE;
		[ FieldInfo( "Control URG", 13 ) ]
		public bool				_controlURG;
		[ FieldInfo( "Control ACK", 13 ) ]
		public bool				_controlACK;
		[ FieldInfo( "Control PSH", 13 ) ]
		public bool				_controlPSH;
		[ FieldInfo( "Control RST", 13 ) ]
		public bool				_controlRST;
		[ FieldInfo( "Control SYN", 13 ) ]
		public bool				_controlSYN;
		[ FieldInfo( "Control FIN", 13 ) ]
		public bool				_controlFIN;
		[ FieldInfo( "Window", 14 ) ]
		public ushort			_window;
		[ FieldInfo( "Check Sum", 16 ) ]
		public ushort			_checkSum;
		[ FieldInfo( "Urgent Pointer", 18 ) ]
		public ushort			_urgentPtr;

		// methods.

		public static new TranProt_TCP GET( RawPacket rawPacket, Packet_Part[] precParts )
		{
			try
			{
				TranProt_TCP		tp = new TranProt_TCP( rawPacket, precParts );
				return tp;
			}
			catch
			{
				return null;
			}
		}
	}

	[ PartInfo( "UDP Transport Protocol header" ) ]
	public class TranProt_UDP : Transport_Protocol
	{
		// construction.

		protected TranProt_UDP( RawPacket rawPacket, Packet_Part[] precParts ) : base()
		{
			// fill the fields.

			int			pos = RawPacket.GetTotalLength( precParts, null, true );

			_srcPort = (ushort) RawPacket.ReadUInt16( rawPacket._data, pos );
			pos += 2;

			_dstPort = (ushort) RawPacket.ReadUInt16( rawPacket._data, pos );
			pos += 2;

			_packetLength = (ushort) RawPacket.ReadUInt16( rawPacket._data, pos );
			pos += 2;

			_checkSum = (ushort) RawPacket.ReadUInt16( rawPacket._data, pos );
			pos += 2;

			// set length.

			_length = 8;
		}

		// data.

		[ FieldInfo( "Source Port", 0 ) ]
		public ushort			_srcPort;
		[ FieldInfo( "Destination Port", 2 ) ]
		public ushort			_dstPort;
		[ FieldInfo( "Packet Length", 4 ) ]
		public ushort			_packetLength;
		[ FieldInfo( "Check Sum", 6 ) ]
		public ushort			_checkSum;

		// methods.

		public static new TranProt_UDP GET( RawPacket rawPacket, Packet_Part[] precParts )
		{
			try
			{
				TranProt_UDP		tp = new TranProt_UDP( rawPacket, precParts );
				return tp;
			}
			catch
			{
				return null;
			}
		}
	}
}
